--

with Bounded_Buffers;
with Printer;

with Ada.Exceptions;
with System.Multiprocessors;

package body Prime_Numbers is
   use Ada;
   use Printer;

   -- -- --

   Done : Boolean := False;
   pragma Atomic (Done);

   Last_Prime : Positive := 2;
   pragma Atomic (Last_Prime);

   Nb_Of_Prime_Pairs : Natural := 0;
   pragma Atomic (Nb_Of_Prime_Pairs);

   Task_Nb : Integer := -1;
   pragma Atomic (Task_Nb);

   My_Number_Of_Cpus : Natural := Natural (System.Multiprocessors.Number_Of_CPUs);
   pragma Atomic (My_Number_Of_Cpus);

   function Set_Cpu return System.Multiprocessors.CPU_Range;

   -- -- --

   package Buffers is new Bounded_Buffers (Item_Type => Natural);

   subtype Source_Buffer_Type is Buffers.Buffer_Type (Buffer_Size => Source_Buffer_Size, Ceiling => Ceiling_Priority);
   subtype Sieve_Buffer_Type is Buffers.Buffer_Type (Buffer_Size => Sieve_Buffer_Size, Ceiling => Ceiling_Priority);

   procedure Test_Buffer_Is_Full (Buf : access Buffers.Buffer_Type; Doit : Boolean := False);

   -- -- --

   task Source_Task is
      pragma Priority (Source_Task_Priority);

      entry Start;
   end Source_Task;

   -- -- --

   task type Sieve_Task_Type
      --
       (Input_Buffer : access Buffers.Buffer_Type; Cpu_Nb : System.Multiprocessors.CPU_Range) with
      Cpu      => Set_Cpu,
      Priority => Sieve_Task_Priority;

   function Get_New_Sieve_Task (In_Buf : access Buffers.Buffer_Type) return access Sieve_Task_Type;

   task body Sieve_Task_Type is separate;

   -- -- --
   -- -- --

   task body Source_Task is
      Limit : constant Positive := Limit_Integer;

      Num : Natural;

      Output_Buffer   : access Source_Buffer_Type := new Source_Buffer_Type;
      Next_Sieve_Task : access Sieve_Task_Type;

   begin
      accept Start;

      Num             := 3;
      Next_Sieve_Task := Get_New_Sieve_Task (In_Buf => Output_Buffer);

      while Num <= Limit loop

         Test_Buffer_Is_Full (Output_Buffer);

         Buffers.Put (Output_Buffer.all, Num);

         Num := Num + 2;
      end loop;

      delay 1.0;

      Num := 0;
      Buffers.Put (Output_Buffer.all, Num);

      delay 1.0;

   end Source_Task;

   -- -- --
   -- -- --

   function Set_Cpu return System.Multiprocessors.CPU_Range is
      use System.Multiprocessors;
   begin
      Task_Nb := Integer'Succ (Task_Nb);

      return CPU_Range (Task_Nb mod My_Number_Of_Cpus + 1);

   exception
      when Constraint_Error =>
         Task_Nb := 0;

         return 1;
   end Set_Cpu;

   -- -- --

   function Get_New_Sieve_Task (In_Buf : access Buffers.Buffer_Type) return access Sieve_Task_Type is
   begin
      return new Sieve_Task_Type (Input_Buffer => In_Buf, Cpu_Nb => Set_Cpu);

   end Get_New_Sieve_Task;

   -- -- --

   procedure Test_Buffer_Is_Full (Buf : access Buffers.Buffer_Type; Doit : Boolean := False) is
   begin
      if Doit and then Buffers.Is_Full (Buf.all) then

         Printer_Task.Print ("#! Buffer is full");
      end if;
   end Test_Buffer_Is_Full;

   -- -- --
   -- -- --

   procedure Start is
   begin
      Source_Task.Start;
   end Start;

end Prime_Numbers;
